//
// Animation player modeless dialog
//
// Author: mla (8-Feb-2020)
//

#include "internal.h"

typedef struct _ANIMDLG_DATA {
    HWND Window;
    const ANIMATION *Animation;
    int CurFrame;
    int Timer;
} ANIMDLG_DATA;

static ANIMDLG_DATA *CreateInstanceData(HWND hwnd, CREATESTRUCT *cs)
{
    ANIMDLG_DATA *self;

    self = malloc(sizeof(ANIMDLG_DATA));
    if(self) {
        memset(self, 0, sizeof(ANIMDLG_DATA));
        self->Window = hwnd;
        self->Animation = (const ANIMATION *) cs->lpCreateParams;
        assert(self->Animation && self->Animation->Hdr.Type == OBJ_ANIMATION);
    }

    return self;
}

static void FreeInstanceData(ANIMDLG_DATA *self)
{
    free(self);
}

//////////////////////////////////////////////////////////////////////////////
// WM_PAINT handler
//////////////////////////////////////////////////////////////////////////////

static void OnPaint(HDC hdc, ANIMDLG_DATA *self)
{
    const ANIMATION *anim = self->Animation;
    const FRAME *frame;
    const IMAGE *image;
    RECT rc;

    GetClientRect(self->Window, &rc);

    if(!anim || !anim->Count || self->CurFrame >= anim->Count) {
        // Clear the window contents
        FillRect(hdc, &rc, (HBRUSH) (COLOR_APPWORKSPACE + 1));
        return;
    }

    frame = anim->Frames[self->CurFrame];
    image = frame->Image;
    viewDrawIMAGE(hdc, 0, 0, rc.right - rc.left, rc.bottom - rc.top,
                  image, image->Palette);
}

//////////////////////////////////////////////////////////////////////////////
// Timer handler
//////////////////////////////////////////////////////////////////////////////

static void OnTimer(ANIMDLG_DATA *self)
{
    const ANIMATION *anim = self->Animation;

    if(!anim || !anim->Count) {
        KillTimer(self->Window, 1);
        InvalidateRect(self->Window, NULL, FALSE);
        UpdateWindow(self->Window);
        return;
    }

    self->Timer--;
    if(self->Timer > 0) {
        return;
    }

    do {
        self->CurFrame++;
        if(self->CurFrame >= anim->Count) self->CurFrame = 0;
    } while(!anim->Frames[self->CurFrame]->Image);

    self->Timer = anim->Frames[self->CurFrame]->Info[1];

    InvalidateRect(self->Window, NULL, FALSE);
    UpdateWindow(self->Window);
}


//////////////////////////////////////////////////////////////////////////////
// Window procedure
//////////////////////////////////////////////////////////////////////////////

static LRESULT CALLBACK
AnimWindowProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    ANIMDLG_DATA *self;
    PAINTSTRUCT ps;

    self = (ANIMDLG_DATA *) GetWindowLongPtr(hwnd, 0);
    switch(message) {
    default:
        break;

    case WM_NCCREATE:
        self = CreateInstanceData(hwnd, (CREATESTRUCT *) lParam);
        if(!self) {
            return FALSE;
        }
        SetWindowLongPtr(hwnd, 0, (LONG_PTR) self);
        break;

    case WM_CREATE:
        SetTimer(hwnd, 1, 1000 / 30, NULL);
        break;

    case WM_PAINT:
        assert(self);
        BeginPaint(hwnd, &ps);
        OnPaint(ps.hdc, self);
        EndPaint(hwnd, &ps);
        break;

    case WM_TIMER:
        assert(self);
        if(wParam == 1) {
            OnTimer(self);
            return 0;
        }
        break;

    case WM_DESTROY:
        appNotifyWindowDestroy(hwnd);
        break;

    case WM_NCDESTROY:
        SetWindowLongPtr(hwnd, 0, 0);
        if(self) {
            FreeInstanceData(self);
            self = NULL;
        }
        break;

    case ANM_SET_ANIM:
        assert(self);
        self->CurFrame = 0;
        self->Timer = 1;
        self->Animation = (const ANIMATION *) lParam;
        KillTimer(hwnd, 1);
        SetTimer(hwnd, 1, 1000 / 30, NULL);
        InvalidateRect(hwnd, NULL, FALSE);
        UpdateWindow(hwnd);
        return 0;
    }

    return DefWindowProcW(hwnd, message, wParam, lParam);
}

//////////////////////////////////////////////////////////////////////////////
// Public functions
//////////////////////////////////////////////////////////////////////////////

void animInit(void)
{
    WNDCLASSEX wc;

    //
    // Register the class
    //

    memset(&wc, 0, sizeof wc);
    wc.cbSize = sizeof wc;
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = AnimWindowProc;
    wc.hInstance = g_AppInstance;
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.cbWndExtra = sizeof(ANIMDLG_DATA *);
    wc.lpszClassName = ANIMDLG_CLASS_NAME;
    if(!RegisterClassExW(&wc)) {
        appWin32Error(GetLastError(), L"RegisterClassExW");
    }
}

HWND animCreate(const ANIMATION *anim)
{
    HWND h;
    RECT rc;
    DWORD style;
    DWORD exStyle;

    // Default size is about four times the size of a normal base image
    SetRect(&rc, 0, 0, 52 * 4, 89 * 4);

    style = WS_OVERLAPPEDWINDOW;
    exStyle = WS_EX_TOOLWINDOW | WS_EX_CLIENTEDGE | WS_EX_TOPMOST;
    AdjustWindowRectEx(&rc, style, FALSE, exStyle);

    h = CreateWindowExW(exStyle,
                        ANIMDLG_CLASS_NAME,
                        L"Animation",
                        style,
                        CW_USEDEFAULT, CW_USEDEFAULT,
                        rc.right - rc.left, rc.bottom - rc.top,
                        NULL, NULL, g_AppInstance, (LPVOID) anim);

    return h;
}
