//
// FACE.DAT loader
//
// Author: mla (7-Feb-2020)
//
//  This loads all the blocks in the FACE.DAT file into the hierarchical
//  structure described in types.h.
//

#include "internal.h"
#include <stddef.h>
#include <setjmp.h>

// If the size in bytes of FACE.DAT is larger than this value,
// then the dbLoad function asks for user confirmation before loading the file.
#define DANGER_THRESHOLD            (16 * 1024 * 1024)

//////////////////////////////////////////////////////////////////////////////
// Utility functions
//////////////////////////////////////////////////////////////////////////////

typedef struct _LDR_CONTEXT {
    HWND ParentWindow;          // for modal dialogs
    HANDLE FileHandle;
    OBJECT *Objects;            // list of all allocated objects
    uint8_t *Buffer;
    uint32_t BufferSize;
    uint32_t HighMark;
    jmp_buf JmpBuf;
} LDR_CONTEXT;

typedef struct _LDR_READER {
    LDR_CONTEXT *Ldr;
    uint32_t Start;
    uint32_t Pos;
    uint32_t End;
} LDR_READER;

// Print to debugger
static void Trace(const WCHAR *format, ...)
{
    va_list ap;
    WCHAR buffer[1024];

    va_start(ap, format);
    StringCbVPrintfW(buffer, sizeof buffer, format, ap);
    va_end(ap);

    OutputDebugStringW(buffer);
}

// Clean up all objects owned by the LDR_CONTEXT
static void ErrorCleanup(LDR_CONTEXT *ldr)
{
    OBJECT *next;

    while(ldr->Objects) {
        next = ldr->Objects->Next;
        free(ldr->Objects);
        ldr->Objects = next;
    }

    assert(!ldr->Objects);

    if(ldr->Buffer) {
        free(ldr->Buffer);
        ldr->Buffer = NULL;
    }
    ldr->BufferSize = 0;

    if(ldr->FileHandle) {
        CloseHandle(ldr->FileHandle);
        ldr->FileHandle = NULL;
    }
}

// Cleanup and return NULL from dbLoad
// The message may contain embedded newlines, but it should not end in one
NORETURN static void Error(LDR_CONTEXT *ldr, const WCHAR *format, ...)
{
    va_list ap;
    WCHAR buffer[1024];

    va_start(ap, format);
    StringCbVPrintfW(buffer, sizeof buffer, format, ap);
    va_end(ap);
    MessageBoxW(ldr->ParentWindow, buffer, NULL, MB_OK | MB_ICONERROR);

    ErrorCleanup(ldr);
    longjmp(ldr->JmpBuf, 1);
}

// Like Error, but called as a result of an error in a Win32 function
NORETURN static void Win32Error(LDR_CONTEXT *ldr, DWORD code, const WCHAR *func)
{
    appWin32Warn(ldr->ParentWindow, code, func);

    ErrorCleanup(ldr);
    longjmp(ldr->JmpBuf, 1);
}

// Allocate and link a new zero-filled object
// Call Error if out of memory
static void *NewObject(LDR_CONTEXT *ldr, uint16_t type, size_t length)
{
    OBJECT *obj;

    assert(length >= sizeof(OBJECT));
    obj = malloc(length);
    if(!obj) {
        Error(ldr, L"Out of memory");
    }

    memset(obj, 0, length);
    obj->Type = type;
    obj->Next = ldr->Objects;
    ldr->Objects = obj;
    return obj;
}

// Allocate an object ``baseLength + elmSize * elmCount`` bytes long
static void *NewVarObj(LDR_CONTEXT *ldr, uint16_t type, size_t baseLength,
                       size_t elmSize, size_t elmCount)
{
    // Check for overflow in `baseLength + elmSize * elmCount`
    if(elmCount >= SIZE_MAX / elmSize ||
       (elmSize * elmCount) >= (SIZE_MAX - baseLength))
    {
        Error(ldr, L"NewVarObj: overflow");
    }

    return NewObject(ldr, type, baseLength + elmSize * elmCount);
}

//////////////////////////////////////////////////////////////////////////////
// Primitive reader functions
//////////////////////////////////////////////////////////////////////////////

static
LDR_READER InitReader(LDR_CONTEXT *ldr, uint32_t start, uint32_t size)
{
    LDR_READER r;

    r.Ldr = ldr;
    r.Start = start;
    r.End = start + size;
    r.Pos = r.Start;

    if(r.Start > r.Ldr->BufferSize || r.End < r.Start || r.End > r.Ldr->BufferSize) {
        Error(ldr, L"InitReader: read overflow");
    }

    return r;
}

// Note ``relOffs`` is relative to ``parent->Start``
static LDR_READER SubReader(const LDR_READER *parent, uint32_t relOffs)
{
    LDR_READER r;

    r.Ldr = parent->Ldr;
    r.Start = parent->Start + relOffs;
    r.End = parent->End;
    r.Pos = r.Start;

    assert(r.End <= r.Ldr->BufferSize);
    return r;
}

static uint8_t ReadU8(LDR_READER *r)
{
    assert(r->End <= r->Ldr->BufferSize);
    if(r->Pos >= r->End) {
        Error(r->Ldr, L"ReadU8: read overflow");
    }

    if(r->Pos > r->Ldr->HighMark) {
        r->Ldr->HighMark = r->Pos;
    }

    assert(r->Pos < r->Ldr->BufferSize);
    return r->Ldr->Buffer[r->Pos++];
}

static uint16_t ReadU16LE(LDR_READER *r)
{
    uint16_t x;
    x = ReadU8(r);
    x |= ReadU8(r) << 8;
    return x;
}

static uint32_t ReadU32LE(LDR_READER *r)
{
    uint32_t x;
    x = ReadU8(r);
    x |= ReadU8(r) << 8;
    x |= ReadU8(r) << 16;
    x |= ReadU8(r) << 24;
    return x;
}

//////////////////////////////////////////////////////////////////////////////
// Palette and image readers
//////////////////////////////////////////////////////////////////////////////

static PALETTE *ReadPalette(LDR_READER r)
{
    PALETTE *palette;
    int i;

    palette = NewObject(r.Ldr, OBJ_PALETTE, sizeof(PALETTE));
    for(i=0; i<256; i++) {
        palette->Entries[i] = ReadU16LE(&r);
    }

    return palette;
}

static IMAGE *ReadImage(LDR_READER r)
{
    unsigned tgtX;
    unsigned tgtY;
    unsigned cx;
    unsigned cy;
    IMAGE *image;
    unsigned stride;
    unsigned x;
    unsigned y;

    // Read image header
    tgtX = ReadU8(&r);
    tgtY = ReadU8(&r);
    cx = ReadU8(&r);
    cy = ReadU8(&r);

    // Allocate image (note stride ensures rows occupy a multiple of 4 bytes)
    stride = (cx + 3U) & ~3U;
    if(!stride) stride = 1;

    image = NewVarObj(r.Ldr, OBJ_IMAGE, offsetof(IMAGE, u.Pixels), stride, cy);
    image->TgtX = tgtX;
    image->TgtY = tgtY;
    image->CX = cx;
    image->CY = cy;

    // Now read the pixels (note rows are not aligned on disk)
    for(y=0; y<cy; y++) {
        for(x=0; x<cx; x++) {
            image->u.Pixels[y * stride + x] = ReadU8(&r);
        }
    }

    return image;
}

//////////////////////////////////////////////////////////////////////////////
// Face reader
//////////////////////////////////////////////////////////////////////////////

static OBJECT *ReadFace(LDR_READER r)
{
    FACE *face;
    uint32_t offsets[8];
    int i;

    // Allocate the face
    face = NewObject(r.Ldr, OBJ_FACE, sizeof(FACE));

    // Read the offsets
    for(i=0; i<8; i++) {
        offsets[i] = ReadU32LE(&r);
    }

    // Read the palette
    if(!offsets[0]) Error(r.Ldr, L"Face has no palette");
    face->Palette = ReadPalette(SubReader(&r, offsets[0]));

    // Read the base image
    if(!offsets[1]) Error(r.Ldr, L"Face has no base image");
    face->Base = ReadImage(SubReader(&r, offsets[1]));
    face->Base->Palette = face->Palette;

    // Now read the variations
    for(i=0; i<NR_VARIATIONS; i++) {
        if(offsets[i + 2]) {
            face->Variations[i] = ReadImage(SubReader(&r, offsets[i + 2]));
            face->Variations[i]->Palette = face->Palette;
            face->Variations[i]->Parent = face->Base;
        }
    }

    return &face->Hdr;
}

//////////////////////////////////////////////////////////////////////////////
// Animation reader
//////////////////////////////////////////////////////////////////////////////

static OBJECT *ReadAnimation(LDR_READER r)
{
    uint32_t frameCount, i;
    ANIMATION *animation;
    FRAME *frame;
    uint32_t palOffs;
    uint32_t imgOffs;
    PALETTE *lastPalette;

    // First read the number of frames
    frameCount = ReadU32LE(&r);

    // Allocate the animation object
    animation = NewVarObj(r.Ldr, OBJ_ANIMATION,
                          offsetof(ANIMATION, Frames),
                          sizeof(FRAME *), frameCount);

    animation->Count = (int) frameCount;

    // Now read the frames
    lastPalette = NULL;
    for(i=0; i<frameCount; i++) {
        frame = NewObject(r.Ldr, OBJ_FRAME, sizeof(FRAME));
        frame->Index = i;
        animation->Frames[i] = frame;

        // Read the frame header
        palOffs = ReadU32LE(&r);
        imgOffs = ReadU32LE(&r);
        frame->Info[0] = ReadU16LE(&r);
        frame->Info[1] = ReadU16LE(&r);

        // Now read the image and the palette, if present
        if(palOffs) {
            frame->Palette = ReadPalette(SubReader(&r, palOffs));
            lastPalette = frame->Palette;
        }

        if(imgOffs) {
            frame->Image = ReadImage(SubReader(&r, imgOffs));
            frame->Image->Palette = lastPalette;
        }
    }

    return &animation->Hdr;
}

//////////////////////////////////////////////////////////////////////////////
// Block reader
//////////////////////////////////////////////////////////////////////////////

static BLOCK *ReadBlock(LDR_READER r)
{
    BLOCK *block;
    uint32_t objCount;
    uint32_t i;
    uint16_t objType;
    uint16_t objId;
    uint32_t objSize;
    uint32_t objOffs;
    OBJECT *obj;
    LDR_READER r2;

    // Do some sanity checks on the block address
    if(r.Start & 2047) Error(r.Ldr, L"Block address is not sector aligned");

    // Read the number of objects
    objCount = ReadU32LE(&r);

    // Allocate the block object
    block = NewVarObj(r.Ldr, OBJ_BLOCK,
                      offsetof(BLOCK, Objects),
                      sizeof(OBJECT *), objCount);

    block->Count = (int) objCount;

    // Now read the subobjects
    for(i=0; i<objCount; i++) {
        objType = ReadU16LE(&r);
        objId = ReadU16LE(&r);
        objSize = ReadU32LE(&r);
        objOffs = ReadU32LE(&r);

        // Note the +4 to skip the objCount field
        r2 = InitReader(r.Ldr, r.Start + 4 + objOffs, objSize);

        switch(objType) {
        case 0:
            obj = ReadFace(r2);
            break;

        case 1:
            obj = ReadAnimation(r2);
            break;

        default:
            Error(r.Ldr, L"Unknown object type 0x%04X", objType);
        }

        // Save the object pointer
        obj->Id = objId;
        block->Objects[i] = obj;
    }

    return block;
}

//////////////////////////////////////////////////////////////////////////////
// File reader
//////////////////////////////////////////////////////////////////////////////

static FACE_DAT *AppendBlock(LDR_CONTEXT *ldr, FACE_DAT *blocks, BLOCK *block)
{
    size_t newSize;
    FACE_DAT *newBlocks;

    newSize = offsetof(FACE_DAT, Blocks) +
                (blocks->Count + 1) * sizeof(BLOCK *);
    newBlocks = realloc(blocks, newSize);
    if(!newBlocks) {
        free(blocks);
        Error(ldr, L"No memory");
    }

    blocks = newBlocks;
    blocks->Blocks[blocks->Count++] = block;
    return blocks;
}

static FACE_DAT *ReadEverything(LDR_CONTEXT *ldr)
{
    LDR_READER r;
    FACE_DAT *blocks;
    BLOCK *block;
    uint32_t next;

    // Allocate an empty block pointer list
    // Note we cannot use NewObject for allocating FACE_DAT since FACE_DAT
    // is not an object
    blocks = malloc(sizeof(FACE_DAT));
    if(!blocks) Error(ldr, L"No memory");
    memset(blocks, 0, sizeof(FACE_DAT));

    r.Ldr = ldr;
    r.Start = 0;
    r.End = ldr->BufferSize;
    ldr->HighMark = 0;

    while(r.Start < r.End) {

        // Skip the block if it is empty
        r.Pos = r.Start;
        if(ReadU32LE(&r)) {     // "peek" at the object count field
            r.Pos = r.Start;
            block = ReadBlock(r);
            block->SectorAddress = r.Start / 2048U;
            block->SectorCount = (ldr->HighMark + 2047U - r.Start) / 2048U;
            blocks = AppendBlock(r.Ldr, blocks, block);
        }

        // Move to the next sector boundary (but check if there is any non-zero
        // bytes in between)
        next = (ldr->HighMark + 2047U) & ~2047U;
        if(next > r.End) next = r.End;

        if(IsDebuggerPresent()) {
            r.Pos = ldr->HighMark + 1;
            while(r.Pos < next) {
                if(ReadU8(&r) != 0) {
                    Trace(L"Nonzero byte at %08x\n", r.Pos - 1);
                }
            }
        }

        r.Start = next;
    }

    return blocks;
}

//////////////////////////////////////////////////////////////////////////////
// Load the raw file bits in
//////////////////////////////////////////////////////////////////////////////

static void LoadFileBits(LDR_CONTEXT *ldr, const WCHAR *path)
{
    HANDLE h;
    LARGE_INTEGER li;
    int r;
    DWORD bytesRead;
    BOOL b;

    // Open the file for read access
    h = CreateFileW(path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
                    FILE_FLAG_SEQUENTIAL_SCAN, NULL);

    if(h == NULL || h == INVALID_HANDLE_VALUE) {
        Win32Error(ldr, GetLastError(), L"CreateFileW");
    }
    ldr->FileHandle = h;

    // Ensure the size is OK
    if(!GetFileSizeEx(h, &li)) {
        Win32Error(ldr, GetLastError(), L"GetFileSizeEx");

    } else if(li.QuadPart >= MAXINT) {
        Error(ldr, L"FACE.DAT file is too large");

    } else if(li.QuadPart & 2047) {
        Error(ldr, L"FACE.DAT file size is not a multiple of 2048");

    } else if(li.QuadPart >= DANGER_THRESHOLD) {
        r = MessageBoxW(ldr->ParentWindow,
                        L"The file seems too large. Load anyway?",
                        NULL,
                        MB_YESNO | MB_ICONWARNING);

        if(r != IDYES) {
            ErrorCleanup(ldr);
            longjmp(ldr->JmpBuf, 2);
        }
    }
    assert(!li.HighPart);

    // Read the file into a memory buffer
    ldr->BufferSize = li.LowPart;
    ldr->Buffer = malloc(ldr->BufferSize);
    if(!ldr->Buffer) {
        Error(ldr, L"No memory for FACE.DAT file buffer");
    }

    b = ReadFile(h,
                 ldr->Buffer,
                 ldr->BufferSize,
                 &bytesRead,
                 NULL);

    if(!b) {
        Win32Error(ldr, GetLastError(), L"ReadFile");

    } else if(bytesRead != ldr->BufferSize) {
        Error(ldr, L"Unexpected EOF while reading file");
    }

    CloseHandle(h);
    ldr->FileHandle = NULL;
}

//////////////////////////////////////////////////////////////////////////////
// Dump functions
//////////////////////////////////////////////////////////////////////////////

static void DumpFace(const FACE *face)
{
    int i;

    Trace(L"  %04X  Face  %dx%d  ", face->Hdr.Id, face->Base->CX, face->Base->CY);
    for(i=0; i<NR_VARIATIONS; i++) {
        if(face->Variations[i]) Trace(L"%d", i);
        else Trace(L" ");
    }
    Trace(L"\n");
}

static void DumpAnimation(const ANIMATION *anim)
{
    FRAME *frame;
    int i;

    Trace(L"  %04X  Animation (%d frames)\n", anim->Hdr.Id, anim->Count);
    for(i=0; i<anim->Count; i++) {
        frame = anim->Frames[i];
        Trace(L"    %2u %s %s %04X %04X\n",
              frame->Index,
              (frame->Image ? L"Img" : L"   "),
              (frame->Palette ? L"Pal" : L"   "),
              frame->Info[0],
              frame->Info[1]);
    }
}

static void DumpBlock(const BLOCK *block)
{
    int i;

    Trace(L"Block %08X [%X] (%u objects)\n", block->SectorAddress, block->SectorCount, block->Count);
    for(i=0; i<block->Count; i++) {
        switch(block->Objects[i]->Type) {
        case OBJ_FACE:
            DumpFace((FACE *) block->Objects[i]);
            break;

        case OBJ_ANIMATION:
            DumpAnimation((ANIMATION *) block->Objects[i]);
            break;

        default:
            Trace(L"  %04X  Unknown object type %04X\n", block->Objects[i]->Id, block->Objects[i]->Type);
            break;
        }
    }
    Trace(L"\n");
}

static void DumpFaceDat(const FACE_DAT *file)
{
    int i;

    Trace(L"Block count: %d\n", file->Count);
    for(i=0; i<file->Count; i++) {
        DumpBlock(file->Blocks[i]);
    }
}

//////////////////////////////////////////////////////////////////////////////
// Public functions
//////////////////////////////////////////////////////////////////////////////

FACE_DAT *dbLoad(HWND parent, const WCHAR *path)
{
    LDR_CONTEXT ldr;
    FACE_DAT *file;

    // Prepare a loader context
    memset(&ldr, 0, sizeof ldr);
    ldr.ParentWindow = parent;

    // Set the return point for Error
    if(setjmp(ldr.JmpBuf)) {
        // Everything cleaned up already by ErrorCleanup
        return NULL;
    }

    // Load the file into a buffer
    LoadFileBits(&ldr, path);

    // Load file contents, transfer ownership of all OBJECTS to the FACE_DAT
    // structure, then cleanup and return
    file = ReadEverything(&ldr);
    file->Objects = ldr.Objects;
    ldr.Objects = NULL;
    ErrorCleanup(&ldr);
    return file;
}

void dbFree(FACE_DAT *file)
{
    OBJECT *p;

    while(file->Objects) {
        p = file->Objects->Next;
        free(file->Objects);
        file->Objects = p;
    }

    free(file);
}