//
// Main program window
//
// Author: mla (7-Feb-2020)
//

#include "internal.h"

#define MAIN_CLASS_NAME     L"FaceView_MainWindow"

// Control IDs
#define CID_SBAR        200     // status bar
#define CID_TBAR        201     // toolbar
#define CID_TVIEW       202     // tree view
#define CID_OBJVIEW     203     // object view

//////////////////////////////////////////////////////////////////////////////
// WM_CREATE handler
//////////////////////////////////////////////////////////////////////////////

static int OnCreate(HWND parent)
{
    static const DWORD style = WS_CHILD | WS_VISIBLE;
    HWND h;

    // Create the object view
    h = CreateWindowExW(WS_EX_CLIENTEDGE,
                        OBJVIEW_CLASS_NAME,
                        NULL,
                        style,
                        0, 0, 0, 0,
                        parent, (HMENU) CID_OBJVIEW, g_AppInstance, NULL);
    if(!h) {
        return 0;
    }

    // Create the tree view
    h = CreateWindowExW(WS_EX_CLIENTEDGE,
                        WC_TREEVIEW,
                        NULL,
                        style | TVS_HASLINES | TVS_LINESATROOT | TVS_HASBUTTONS,
                        0, 0, 0, 0,
                        parent, (HMENU) CID_TVIEW, g_AppInstance, NULL);
    if(!h) {
        return 0;
    }

    // Create the status bar
    h = CreateWindowExW(0,
                        STATUSCLASSNAME,
                        NULL,
                        style | SBARS_SIZEGRIP,
                        0, 0, 0, 0,
                        parent, (HMENU) CID_SBAR, g_AppInstance, NULL);
    if(!h) {
        return 0;
    }

    // Create the toolbar
    h = CreateWindowExW(0,
                        TOOLBARCLASSNAME,
                        NULL,
                        style | TBSTYLE_WRAPABLE,
                        0, 0, 0, 0,
                        parent, (HMENU) CID_TBAR, g_AppInstance, NULL);
    if(!h) {
        return 0;
    }
    SendMessageW(h, TB_BUTTONSTRUCTSIZE, (WPARAM) sizeof(TBBUTTON), 0);

    // Create the toolbar buttons

    // TODO:

#define imageListID 0
    HIMAGELIST imageList;

    TBBUTTON tbButtons[] = {
        { MAKELONG(STD_FILEOPEN, imageListID), IDM_LOAD,
          TBSTATE_ENABLED, BTNS_AUTOSIZE,
          {0}, 0, 0 },
    };

    imageList = ImageList_Create(16, 16, ILC_COLOR16 | ILC_MASK, ARRAY_SIZE(tbButtons), 0);
    SendMessage(h, TB_SETIMAGELIST, (WPARAM) imageListID, (LPARAM) imageList);
    SendMessage(h, TB_LOADIMAGES, (WPARAM) IDB_STD_SMALL_COLOR, (LPARAM) HINST_COMMCTRL);
    SendMessage(h, TB_ADDBUTTONS, ARRAY_SIZE(tbButtons), (LPARAM) tbButtons);
    SendMessage(h, TB_AUTOSIZE, 0, 0);

#undef imageListID

    return 1;
}

//////////////////////////////////////////////////////////////////////////////
// Tree fill code
//////////////////////////////////////////////////////////////////////////////

static HTREEITEM
InsertItem(HWND tree, HTREEITEM anchor, const void *ptr, const WCHAR *label)
{
    TVINSERTSTRUCT tvi;

    memset(&tvi, 0, sizeof tvi);
    tvi.hParent = anchor;
    tvi.hInsertAfter = TVI_LAST;
    tvi.item.mask = TVIF_PARAM | TVIF_TEXT;
    tvi.item.lParam = (LPARAM) ptr;
    tvi.item.pszText = (WCHAR *) label;
    return (HTREEITEM) SendMessageW(tree, TVM_INSERTITEM, 0, (LPARAM) &tvi);
}

static void
InsertImage(HWND tree, HTREEITEM anchor, const IMAGE *image, const WCHAR *label)
{
    InsertItem(tree, anchor, image, label);
}

static void InsertPalette(HWND tree, HTREEITEM anchor, const PALETTE *palette)
{
    InsertItem(tree, anchor, palette, L"Palette");
}

static void InsertFace(HWND tree, HTREEITEM anchor, const FACE *face)
{
    static const WCHAR *VariationNames[NR_VARIATIONS] = {
        L"Eyes1",
        L"Eyes2",
        L"NotUsed_2",
        L"Mouth1",
        L"Mouth2",
        L"NotUsed_5"
    };

    WCHAR label[256];
    HTREEITEM root;
    int i;

    StringCbPrintf(label, sizeof label, L"Face %04X", face->Hdr.Id);

    root = InsertItem(tree, anchor, face, label);
    if(face->Palette) InsertPalette(tree, root, face->Palette);
    if(face->Base) InsertImage(tree, root, face->Base, L"Base");
    for(i=0; i<NR_VARIATIONS; i++) {
        if(face->Variations[i]) {
            InsertImage(tree, root, face->Variations[i], VariationNames[i]);
        }
    }
}

static void InsertFrame(HWND tree, HTREEITEM anchor, const FRAME *frame)
{
    WCHAR label[256];
    HTREEITEM root;

    StringCbPrintf(label, sizeof label, L"Frame %u (%04X %04X)",
                   frame->Index, frame->Info[0], frame->Info[1]);

    root = InsertItem(tree, anchor, frame, label);
    if(frame->Image) InsertImage(tree, root, frame->Image, L"Pixels");
    if(frame->Palette) InsertPalette(tree, root, frame->Palette);
}

static void InsertAnimation(HWND tree, HTREEITEM anchor, const ANIMATION *anim)
{
    WCHAR label[256];
    HTREEITEM root;
    int i;

    StringCbPrintfW(label, sizeof label, L"Animation %04X (%d frame%s)",
                    anim->Hdr.Id,
                    anim->Count,
                    (anim->Count == 1 ? "" : "s"));

    root = InsertItem(tree, anchor, anim, label);
    for(i=0; i<anim->Count; i++) {
        InsertFrame(tree, root, anim->Frames[i]);
    }
}

static void InsertBlock(HWND tree, const BLOCK *block)
{
    WCHAR label[256];
    HTREEITEM root;
    int i;

    StringCbPrintfW(label, sizeof label, L"Block %u [%u]", block->SectorAddress, block->SectorCount);

    root = InsertItem(tree, TVI_ROOT, block, label);
    for(i=0; i<block->Count; i++) {
        switch(block->Objects[i]->Type) {
        default:
            FatalAppExitW(0, L"Invalid object type");

        case OBJ_FACE:
            InsertFace(tree, root, (FACE *) block->Objects[i]);
            break;

        case OBJ_ANIMATION:
            InsertAnimation(tree, root, (ANIMATION *) block->Objects[i]);
            break;
        }
    }
}

static void PopulateTree(HWND tree, const FACE_DAT *file)
{
    int i;

    for(i=0; i<file->Count; i++) {
        InsertBlock(tree, file->Blocks[i]);
    }
}

//////////////////////////////////////////////////////////////////////////////
// WM_COMMAND(IDM_LOAD) handler
//////////////////////////////////////////////////////////////////////////////

// TODO: make instance data
static FACE_DAT *g_Data;
static OBJECT *g_CurObj;
static HWND g_Anim;

static void OnLoad(HWND parent)
{
    static WCHAR pathBuf[32768]; // static so it's preserved across calls
    OPENFILENAMEW ofn;
    FACE_DAT *newData;

    memset(&ofn, 0, sizeof ofn);
    ofn.lStructSize = sizeof ofn;
    ofn.hwndOwner = parent;
    ofn.lpstrFilter = L"Data Files (*.dat)\0*.DAT\0"
                      L"All Files (*.*)\0*.*\0"
                      L"\0\0";
    ofn.lpstrFile = pathBuf;
    ofn.nMaxFile = ARRAY_SIZE(pathBuf);
    ofn.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
    ofn.lpstrDefExt = L"DAT";
    if(!GetOpenFileNameW(&ofn)) {
        return;
    }

    // Load the new data before changing anything
    newData = dbLoad(parent, pathBuf);
    if(!newData) {
        return;
    }

    TreeView_DeleteAllItems(GetDlgItem(parent, CID_TVIEW));
    SendMessage(GetDlgItem(parent, CID_OBJVIEW), OVM_RESET, 0, 0);
    if(g_Anim) {
        DestroyWindow(g_Anim);
        g_Anim = NULL;
    }
    g_CurObj = NULL;

    if(g_Data) dbFree(g_Data);
    g_Data = newData;

    PopulateTree(GetDlgItem(parent, CID_TVIEW), g_Data);
}

//////////////////////////////////////////////////////////////////////////////
// WM_SIZE handler
//////////////////////////////////////////////////////////////////////////////

static void OnSize(HWND parent)
{
    RECT rc;
    int tviewCX;

    // Calculate the window portion after removing the status bar and the
    // tool bar (note this is called after forwarding WM_SIZE to the status
    // and tool bars)

    GetClientRect(parent, &rc);
    {
        RECT rc2;

        GetWindowRect(GetDlgItem(parent, CID_TBAR), &rc2);
        rc.top += rc2.bottom - rc2.top;

        GetWindowRect(GetDlgItem(parent, CID_SBAR), &rc2);
        rc.bottom -= rc2.bottom - rc2.top;
    }

    // Move the tree view

    tviewCX = 224;  // TODO: make movable

    MoveWindow(GetDlgItem(parent, CID_TVIEW),
               rc.left, rc.top,
               tviewCX, rc.bottom - rc.top,
               TRUE);

    // Move the object view

    MoveWindow(GetDlgItem(parent, CID_OBJVIEW),
               rc.left + tviewCX, rc.top,
               (rc.right - rc.left) - tviewCX, rc.bottom - rc.top,
               TRUE);
}

//////////////////////////////////////////////////////////////////////////////
// Window procedure
//////////////////////////////////////////////////////////////////////////////

static LRESULT CALLBACK
MainWindowProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message) {

    case APP_WINDOW_DESTROYED:
        if(lParam == (LPARAM) g_Anim) {
            g_Anim = NULL;
        }
        return 0;

    case WM_NOTIFY:
        if(wParam == CID_TVIEW) {
            NMHDR *nm = (NMHDR *) lParam;

            if(nm->code == TVN_SELCHANGED) {
                NMTREEVIEW *nt = (NMTREEVIEW *) lParam;

                // Change the object shown in the object viewer

                g_CurObj = (OBJECT *) nt->itemNew.lParam;
                SendMessage(GetDlgItem(hwnd, CID_OBJVIEW),
                            OVM_SET_OBJECT,
                            0,
                            nt->itemNew.lParam);

            } else if(nm->code == TVN_KEYDOWN) {
                NMTVKEYDOWN *vk = (NMTVKEYDOWN *) lParam;

                // Open the animation window

                if(vk->wVKey == VK_RETURN) {
                    if(g_CurObj && g_CurObj->Type == OBJ_ANIMATION) {
                        if(!g_Anim) {
                            g_Anim = animCreate((ANIMATION*) g_CurObj);
                        } else {
                            SendMessage(g_Anim, ANM_SET_ANIM, 0, (LPARAM) g_CurObj);
                        }
                        ShowWindow(g_Anim, SW_SHOW);
                    }
                }
            }
        }
        break;

    case WM_CREATE:
        appSubscribeWindowDestroy(hwnd);
        if(!OnCreate(hwnd)) {
            return -1;
        }
        return 0;

    case WM_COMMAND:
        if(LOWORD(wParam) == IDM_LOAD) {
            OnLoad(hwnd);
        } else if(LOWORD(wParam) == IDM_EXIT) {
            DestroyWindow(hwnd);
        }
        return 0;

    case WM_DESTROY:
        appUnsubscribeWindowDestroy(hwnd);
        appNotifyWindowDestroy(hwnd);
        PostQuitMessage(0);
        break;

    case WM_SIZE:
        SendMessage(GetDlgItem(hwnd, CID_SBAR), WM_SIZE, wParam, lParam);
        SendMessage(GetDlgItem(hwnd, CID_TBAR), WM_SIZE, wParam, lParam);
        OnSize(hwnd);
        return 0;

    case WM_SETFOCUS:
        SetFocus(GetDlgItem(hwnd, CID_TVIEW));
        return 0;

    default:
        break;
    }

    return DefWindowProcW(hwnd, message, wParam, lParam);
}

//////////////////////////////////////////////////////////////////////////////
// Public functions
//////////////////////////////////////////////////////////////////////////////

void mainInit(void)
{
    WNDCLASSEX wc;

    memset(&wc, 0, sizeof wc);
    wc.cbSize = sizeof wc;
    wc.style = 0;
    wc.lpfnWndProc = MainWindowProc;
    wc.hInstance = g_AppInstance;
    wc.hIcon = LoadIcon(g_AppInstance, L"appIcon");
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.lpszClassName = MAIN_CLASS_NAME;
    wc.lpszMenuName = L"appMenu";
    if(!RegisterClassExW(&wc)) {
        appWin32Error(GetLastError(), L"RegisterClassExW");
    }
}

HWND mainCreateWindow(void)
{
    HWND h;
    DWORD style;
    DWORD exStyle;

    style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    exStyle = 0;

    h = CreateWindowExW(exStyle,
                        MAIN_CLASS_NAME,
                        L"FaceView",
                        style,
                        CW_USEDEFAULT, CW_USEDEFAULT, 500, 400,
                        NULL, NULL, g_AppInstance, NULL);
    if(!h) {
        appWin32Error(GetLastError(), L"CreateWindowExW");
    }

    return h;
}
