//
// OBJECT viewer control
//
// Author: mla (8-Feb-2020)
//

#include "internal.h"
#include <stddef.h>

// A BITMAPINFO with a 256-color palette
typedef struct _BITMAPINFO_256 {
    BITMAPINFOHEADER bmiHeader;
    RGBQUAD bmiColors[256];
} BITMAPINFO_256;

// A 16x16 IMAGE that is used for drawing palettes
typedef union _IMAGE_16x16 {
    IMAGE Img;
    struct {
        uint8_t Header[offsetof(IMAGE, u.Pixels)];
        uint8_t Pixels[16 * 16];
    } u;
} IMAGE_16x16;

static HFONT g_Font;

//////////////////////////////////////////////////////////////////////////////
// Instance data
//////////////////////////////////////////////////////////////////////////////

typedef struct _OBJVIEW_DATA {
    HWND Window;
    OBJECT *Current;
} OBJVIEW_DATA;

static OBJVIEW_DATA *CreateInstanceData(HWND hwnd)
{
    OBJVIEW_DATA *self;

    self = malloc(sizeof(OBJVIEW_DATA));
    if(self) {
        memset(self, 0, sizeof(OBJVIEW_DATA));
        self->Window = hwnd;
    }

    return self;
}

static void FreeInstanceData(OBJVIEW_DATA *self)
{
    free(self);
}

//////////////////////////////////////////////////////////////////////////////
// Draw an IMAGE at dx,dy, stretched to cx,cy
//////////////////////////////////////////////////////////////////////////////

void
viewDrawIMAGE(HDC hdc, int dx, int dy, int cx, int cy,
              const IMAGE *image, const PALETTE *palette)
{
    BITMAPINFO_256 bi;
    int i;

    //
    // Prepare the BITMAPINFO for the stretch blit below
    //

    memset(&bi, 0, sizeof bi);
    bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bi.bmiHeader.biWidth = image->CX;
    bi.bmiHeader.biHeight = -image->CY;
    bi.bmiHeader.biPlanes = 1;
    bi.bmiHeader.biBitCount = 8;
    bi.bmiHeader.biCompression = BI_RGB;
    bi.bmiHeader.biSizeImage = ((image->CX + 3) & ~3) * image->CY;

    if(palette) {
        for(i=0; i<256; i++) {
            uint16_t c = palette->Entries[i];
            bi.bmiColors[i].rgbRed   = (BYTE) (((c >>  0) & 0x1F) << (8 - 5));
            bi.bmiColors[i].rgbGreen = (BYTE) (((c >>  5) & 0x1F) << (8 - 5));
            bi.bmiColors[i].rgbBlue  = (BYTE) (((c >> 10) & 0x1F) << (8 - 5));
        }
    } else {
        for(i=0; i<256; i++) {
            bi.bmiColors[i].rgbRed   = (BYTE) i;
            bi.bmiColors[i].rgbGreen = (BYTE) i;
            bi.bmiColors[i].rgbBlue  = (BYTE) i;
        }
    }

    //
    // Now do the stretch blit
    //

    StretchDIBits(hdc,
                  dx, dy, cx, cy,
                  0, 0, image->CX, image->CY,
                  image->u.Pixels, (BITMAPINFO *) &bi,
                  DIB_RGB_COLORS, SRCCOPY);
}

//////////////////////////////////////////////////////////////////////////////
// Paint handler (default handler)
//////////////////////////////////////////////////////////////////////////////

static void OnPaint_Default(HDC hdc, OBJVIEW_DATA *self)
{
    static const char *TypeNames[] = {
        "Face", "Animation", "Block", "Image", "Palette", "Frame"
    };

    OBJECT *obj = self->Current;
    char buf[256];

    if(obj && (obj->Type == OBJ_FACE || obj->Type == OBJ_ANIMATION)) {
        StringCbPrintfA(buf, sizeof buf, "%s 0x%04X @ %p",
                        TypeNames[obj->Type], obj->Id, obj);
    } else if(obj) {
        StringCbPrintfA(buf, sizeof buf, "%s @ %p",
                        TypeNames[obj->Type], obj);
    } else {
        StringCbCopyA(buf, sizeof buf, "No object selected");
    }

    TextOutA(hdc, 8, 8, buf, (int) strlen(buf));
}

//////////////////////////////////////////////////////////////////////////////
// Paint handler (PALETTE)
//////////////////////////////////////////////////////////////////////////////

static void OnPaint_PALETTE(HDC hdc, const PALETTE *palette)
{
    int imgX;
    int imgY;
    IMAGE_16x16 image;
    int i;

    assert(palette->Hdr.Type == OBJ_PALETTE);

    // Print the label
    TextOutA(hdc, 8, 8, "Palette", (int) strlen("Palette"));

    // Construct a 16x16 image which contains all palette colors
    memset(&image, 0, sizeof image);
    image.Img.CX = 16;
    image.Img.CY = 16;
    for(i=0; i<16*16; i++) {
        image.u.Pixels[i] = (uint8_t)i;
    }

    // Draw the image
    imgX = 8;
    imgY = 8 + 20;
    viewDrawIMAGE(hdc,
                  imgX, imgY,
                  5 * image.Img.CX, 5 * image.Img.CY, &image.Img,
                  palette);
}

//////////////////////////////////////////////////////////////////////////////
// Paint handler (IMAGE)
//////////////////////////////////////////////////////////////////////////////

static void OnPaint_IMAGE(HDC hdc, const IMAGE *image)
{
    char buf[256];
    int imgX;
    int imgY;
    const IMAGE *p;

    assert(image->Hdr.Type == OBJ_IMAGE);

    // Print the image dimensions and target coordinates.
    if(image->TgtX || image->TgtX) {
        StringCbPrintfA(buf, sizeof buf, "Image (%dx%d pixels) at %d,%d",
                        image->CX, image->CY, image->TgtX, image->TgtY);
    } else {
        StringCbPrintfA(buf, sizeof buf, "Image (%dx%d pixels)",
                        image->CX, image->CY);
    }
    TextOutA(hdc, 8, 8, buf, (int) strlen(buf));

    // Draw the image
    imgX = 8;
    imgY = 8 + 20;
    viewDrawIMAGE(hdc, imgX, imgY, image->CX, image->CY, image, image->Palette);
    viewDrawIMAGE(hdc, imgX + image->CX, imgY, image->CX * 2, image->CY * 2, image, image->Palette);

    // If this is a variation image, then render it mixed with the base image.

    if(image->Parent) {
        p = image->Parent;

        imgY += 16 + image->CY * 2;
        StringCbCopyA(buf, sizeof buf, "Mixed with base image:");
        TextOutA(hdc, imgX, imgY, buf, (int) strlen(buf));
        imgY += 24;

        viewDrawIMAGE(hdc, imgX, imgY, p->CX, p->CY, p, p->Palette);
        viewDrawIMAGE(hdc,
                      imgX + image->TgtX, imgY + image->TgtY,
                      image->CX, image->CY, image, image->Palette);

        imgX += p->CX;
        viewDrawIMAGE(hdc, imgX, imgY, 2 * p->CX, 2 * p->CY, p, p->Palette);
        viewDrawIMAGE(hdc,
                      imgX + 2 * image->TgtX, imgY + 2 * image->TgtY,
                      2 * image->CX, 2 * image->CY, image, image->Palette);
    }
}

//////////////////////////////////////////////////////////////////////////////
// Paint handler (FRAME)
//////////////////////////////////////////////////////////////////////////////

static void OnPaint_FRAME(HDC hdc, const FRAME *frame)
{
    char buf[256];
    int imgX;
    int imgY;
    const IMAGE *image;

    assert(frame->Hdr.Type == OBJ_FRAME);

    // Print the frame label

    StringCbPrintfA(buf, sizeof buf, "Frame %-2u (%04X - %04X)",
                    frame->Index, frame->Info[0], frame->Info[1]);
    TextOutA(hdc, 8, 8, buf, (int) strlen(buf));

    // Render the frame (note we use the palette in frame->Image->Palette,
    // not the frame->Palette)

    image = frame->Image;
    imgX = 8;
    imgY = 8 + 20;

    if(image) {
        viewDrawIMAGE(hdc, imgX, imgY, image->CX, image->CY, image, image->Palette);
        viewDrawIMAGE(hdc, imgX + image->CX, imgY, 2 * image->CX, 2 * image->CY, image, image->Palette);

    } else {
        StringCbCopyA(buf, sizeof buf, "No image data");
        TextOutA(hdc, imgX, imgY, buf, (int) strlen(buf));
    }
}

//////////////////////////////////////////////////////////////////////////////
// Paint handler (BLOCK)
//////////////////////////////////////////////////////////////////////////////

static void OnPaint_BLOCK(HDC hdc, OBJVIEW_DATA *self, const BLOCK *block)
{
    HWND hwnd = self->Window;
    RECT rc;
    char buf[256];
    int i;
    int imgX0;
    int imgX;
    int imgY;
    const IMAGE *image;
    int j;
    HGDIOBJ prevBrush;
    HGDIOBJ prevPen;

    assert(block->Hdr.Type == OBJ_BLOCK);

    GetClientRect(hwnd, &rc);

    // Print the frame label

    StringCbPrintfA(buf, sizeof buf, "Block (address %u, size %u), %d object%s",
                    block->SectorAddress, block->SectorCount,
                    block->Count,
                    (block->Count == 1 ? "" : "s"));
    TextOutA(hdc, 8, 8, buf, (int) strlen(buf));

    // Blit a single frame out of all objects into a 2D arrangement

    imgX0 = 8;
    imgX = imgX0;
    imgY = 8 + 20;

    // Prepare for drawing a red rectangle around animations
    prevPen = SelectObject(hdc, GetStockObject(DC_PEN));
    prevBrush = SelectObject(hdc, GetStockObject(NULL_BRUSH));
    SetDCPenColor(hdc, RGB(0xFF, 0, 0));

    for(i=0; i<block->Count; i++) {

        image = NULL;
        if(block->Objects[i]->Type == OBJ_FACE) {
            const FACE *p = (const FACE *) block->Objects[i];
            image = p->Base;

        } else if(block->Objects[i]->Type == OBJ_ANIMATION) {
            const ANIMATION *p = (const ANIMATION *) block->Objects[i];

            // Pick the image from the first frame that has an image
            // (frame images are optional)
            for(j=0; j<p->Count && !image; j++) {
                image = p->Frames[j]->Image;
            }
        }

        if(!image) continue;

        if(imgX + image->CX >= rc.right) {
            imgX = imgX0;
            imgY += image->CY;
        }

        viewDrawIMAGE(hdc, imgX, imgY, image->CX, image->CY, image, image->Palette);

        // Draw a red rectangle around animations
        if(block->Objects[i]->Type == OBJ_ANIMATION) {
            Rectangle(hdc, imgX, imgY, imgX+image->CX, imgY+image->CY);
        }

        imgX += image->CX;
    }

    SelectObject(hdc, prevPen);
    SelectObject(hdc, prevBrush);
}

//////////////////////////////////////////////////////////////////////////////
// Paint handler (FACE)
//////////////////////////////////////////////////////////////////////////////

static void OnPaint_FACE(HDC hdc, const FACE *face)
{
    static const WCHAR *VariationNames[NR_VARIATIONS] = {
        L"Eyes1",
        L"Eyes2",
        L"NotUsed_2",
        L"Mouth1",
        L"Mouth2",
        L"NotUsed_5"
    };

    char buf[256];
    int i;
    int imgX0;
    int imgX;
    int imgY;
    const IMAGE *image;
    const IMAGE *var;

    assert(face->Hdr.Type == OBJ_FACE);

    // Print the label

    StringCbPrintfA(buf, sizeof buf, "Face %04X", face->Hdr.Id);
    TextOutA(hdc, 8, 8, buf, (int) strlen(buf));

    // Blit the base image plus the base image mixed with each variation
    imgX0 = 8;
    imgX = imgX0;
    imgY = 8 + 20;

    image = face->Base;
    viewDrawIMAGE(hdc, imgX, imgY, image->CX * 2, image->CY * 2, image, image->Palette);
    imgX += image->CX * 2 + 16;

    for(i=0; i<NR_VARIATIONS; i++) {
        var = face->Variations[i];
        if(!var) continue;

        viewDrawIMAGE(hdc, imgX, imgY, image->CX * 2, image->CY * 2, image, image->Palette);
        viewDrawIMAGE(hdc, imgX + var->TgtX * 2, imgY + var->TgtY * 2, var->CX * 2, var->CY * 2, var, var->Palette);

        TextOutW(hdc, imgX, imgY + image->CY * 2 + 4,
                 VariationNames[i], (int) wcslen(VariationNames[i]));

        imgX += image->CX * 2;
    }
}

//////////////////////////////////////////////////////////////////////////////
// Paint handler (ANIMATION)
//////////////////////////////////////////////////////////////////////////////

static void OnPaint_ANIMATION(HDC hdc, const ANIMATION *anim)
{
    WCHAR buf[256];
    int i;
    int imgX0;
    int imgX;
    int imgY;
    const IMAGE *image;

    assert(anim->Hdr.Type == OBJ_ANIMATION);

    // Print the label

    StringCbPrintfW(buf, sizeof buf, L"Animation %04X (%d frame%s) - Press ENTER to play",
                    anim->Hdr.Id,
                    anim->Count,
                    (anim->Count == 1 ? L"" : L"s"));
    TextOutW(hdc, 8, 8, buf, (int) wcslen(buf));

    // Blit all frames

    imgX0 = 8;
    imgX = imgX0;
    imgY = 8 + 20;

    for(i=0; i<anim->Count; i++) {
        image = anim->Frames[i]->Image;
        if(!image) continue;

        viewDrawIMAGE(hdc, imgX, imgY, image->CX, image->CY, image, image->Palette);
        StringCbPrintfW(buf, sizeof buf, L"%u", anim->Frames[i]->Index);
        TextOutW(hdc, imgX, imgY + image->CY + 4, buf, (int) wcslen(buf));
        imgX += image->CX;
    }
}

//////////////////////////////////////////////////////////////////////////////
// WM_PAINT handler
//////////////////////////////////////////////////////////////////////////////

static void OnPaint(OBJVIEW_DATA *self)
{
    HWND hwnd = self->Window;
    PAINTSTRUCT ps;
    RECT rc;
    HFONT prevFont;

    // Get a DC handle and setup common DC parameters

    BeginPaint(hwnd, &ps);
    GetClientRect(hwnd, &rc);
    FillRect(ps.hdc, &rc, GetStockObject(WHITE_BRUSH));
    prevFont = SelectFont(ps.hdc, g_Font);

    if(!self->Current) {
        OnPaint_Default(ps.hdc, self);

    } else {
        switch(self->Current->Type) {
        default:
            OnPaint_Default(ps.hdc, self);
            break;

        case OBJ_PALETTE:
            OnPaint_PALETTE(ps.hdc, (const PALETTE *) self->Current);
            break;

        case OBJ_IMAGE:
            OnPaint_IMAGE(ps.hdc, (const IMAGE *) self->Current);
            break;

        case OBJ_FRAME:
            OnPaint_FRAME(ps.hdc, (const FRAME *) self->Current);
            break;

        case OBJ_BLOCK:
            OnPaint_BLOCK(ps.hdc, self, (const BLOCK *) self->Current);
            break;

        case OBJ_FACE:
            OnPaint_FACE(ps.hdc, (const FACE *) self->Current);
            break;

        case OBJ_ANIMATION:
            OnPaint_ANIMATION(ps.hdc, (const ANIMATION *) self->Current);
            break;
        }
    }

    // Restore DC state and finish painting

    SelectFont(ps.hdc, prevFont);
    EndPaint(hwnd, &ps);
}

//////////////////////////////////////////////////////////////////////////////
// Window procedure
//////////////////////////////////////////////////////////////////////////////

static LRESULT CALLBACK
ViewWindowProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    OBJVIEW_DATA *self;

    self = (OBJVIEW_DATA *) GetWindowLongPtr(hwnd, 0);
    switch(message) {

    case WM_DESTROY:
        appNotifyWindowDestroy(hwnd);
        break;

    case WM_NCCREATE:
        self = CreateInstanceData(hwnd);
        if(!self) {
            return FALSE;
        }
        SetWindowLongPtr(hwnd, 0, (LONG_PTR) self);
        break;

    case WM_NCDESTROY:
        SetWindowLongPtr(hwnd, 0, 0);
        if(self) {
            FreeInstanceData(self);
            self = NULL;
        }
        break;

    case OVM_RESET:
        assert(self);
        self->Current = NULL;
        InvalidateRect(hwnd, NULL, TRUE);
        return 0;

    case OVM_SET_OBJECT:
        assert(self);
        self->Current = (OBJECT *) lParam;
        InvalidateRect(hwnd, NULL, TRUE);
        return 0;

    case WM_PAINT:
        assert(self);
        OnPaint(self);
        break;

    default:
        break;
    }

    return DefWindowProcW(hwnd, message, wParam, lParam);
}

//////////////////////////////////////////////////////////////////////////////
// Public functions
//////////////////////////////////////////////////////////////////////////////

void viewInit(void)
{
    LOGFONTW lf;
    WNDCLASSEX wc;

    // Register the class
    memset(&wc, 0, sizeof wc);
    wc.cbSize = sizeof wc;
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = ViewWindowProc;
    wc.hInstance = g_AppInstance;
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.cbWndExtra = sizeof(OBJVIEW_DATA *);
    wc.lpszClassName = OBJVIEW_CLASS_NAME;
    if(!RegisterClassExW(&wc)) {
        appWin32Error(GetLastError(), L"RegisterClassExW");
    }

    // Create the fixed font (ignore errors)
    memset(&lf, 0, sizeof lf);
    lf.lfHeight = -14;
    lf.lfPitchAndFamily = FIXED_PITCH;
    g_Font = CreateFontIndirectW(&lf);
    if(!g_Font) {
        g_Font = GetStockObject(ANSI_FIXED_FONT);
    }
}
