//
// FACE.DAT types
//
// Author: mla (7-Feb-2020)
//
//  The FACE.DAT file is a concatenation of blocks. A block starts on
//  a CD sector boundary (2048 bytes). The radio script header contains
//  the sector address of the face block that is used in that particular script,
//  and the face block contains just the faces for the characters that appear
//  in that particular script.
//
//  The radio script refers to face blocks directly using a sector address;
//  the FACE.DAT file has no "directory"/"block table"/...
//
//  A block contains a series of objects. Every object is identified by
//  a 16-bit ID, which may be passed to the radio script commands
//  (e.g. show face 0xF123). There are two types of objects:
//  faces and animations.
//
//  Faces
//  =====
//
//  A face contains a palette, a base image, and a set of up to 6 variations
//  for "eyes closed", "mouth open", etc. The face is rendered by blitting
//  the base image, then overlaying any active variations: the header for
//  each variation image contains the target position where it is blitted into,
//  relative to the base image.
//
//  The variations are automatically selected by the VOX stream lip sync data
//  (e.g. the radio script does not actually select the variations). Note that
//  most faces do not have all 6 variations (variations are optional). However,
//  the variation index to subimage mapping is "fixed": e.g. variations
//  #0 and #1 control the eyes, variations #3 and #4 control the mouth, etc...
//
//  Animations
//  ==========
//
//  An animation contains a sequence of frames. Every frame contains
//  an optional image, an optional palette, and frame timing info. Some frames
//  do not have neither palette nor image: these are used only for the frame
//  delay.
//

#pragma once
#include <stdint.h>

//////////////////////////////////////////////////////////////////////////////
// In-memory types
//////////////////////////////////////////////////////////////////////////////

//
// Object allocation header. The objects are threaded together so they can
// be cleaned up easily without having to traverse the whole structure
//

typedef struct _OBJECT {
    uint16_t Id;                // object ID (for face and animation only)
    uint16_t Type;
    struct _OBJECT *Next;
} OBJECT;

// OBJECT.Type values
#define OBJ_FACE            0
#define OBJ_ANIMATION       1
#define OBJ_BLOCK           2
#define OBJ_IMAGE           3
#define OBJ_PALETTE         4
#define OBJ_FRAME           5

//
// A palette is an ordered sequence of 256 colors. Each color is RGB555
// (lowest 5 bits red, next 5 bits green, next 5 bits blue).
//

typedef struct _PALETTE {
    OBJECT Hdr;
    uint16_t Entries[256];
} PALETTE;

//
// An image is a 2D array of pixels plus a target position.
//
// Note the Pixels array contains the pixel data, as a row-major 2D array
// of palette indices. Each row starts on a 4-byte boundary. This is done
// so that a pointer to Pixels can be passed directly to functions
// like StretchDIBits (the actual on-disk bits are not necessarily aligned)
//

typedef struct _IMAGE {
    OBJECT Hdr;
    PALETTE *Palette;           // for display purposes only
    struct _IMAGE *Parent;      // overlay parent (display only)
    int CX, CY;                 // dimensions
    int TgtX, TgtY;             // target position
    union {
        uint32_t Alignment;
        uint8_t Pixels[1];      // data (8 bpp, rows aligned to 4 byte boundary)
    } u;
} IMAGE;

//
// Variation indices
//
// Note that, generally, the base image has eyes open and mouth closed.
//

#define VAR_Eyes1           0
#define VAR_Eyes2           1
#define VAR_NotUsed_2       2
#define VAR_Mouth1          3
#define VAR_Mouth2          4
#define VAR_NotUsed_5       5

#define NR_VARIATIONS       6

//
// A face contains a palette, a base image, and several variations.
//

typedef struct _FACE {
    OBJECT Hdr;                         // object header
    PALETTE *Palette;                   // palette
    IMAGE *Base;                        // base image
    IMAGE *Variations[NR_VARIATIONS];   // variations (entries are optional)
} FACE;

//
// A frame is a 2D image, a palette, and some info
//

typedef struct _FRAME {
    OBJECT Hdr;
    uint32_t Index;                     // zero-based
    IMAGE *Image;                       // the frame data (optional)
    PALETTE *Palette;                   // the palette (optional)
    uint16_t Info[2];                   // TODO: format unknown
} FRAME;

//
// An animation is a sequence of frames.
//

typedef struct _ANIMATION {
    OBJECT Hdr;                         // object header
    int Count;                          // number of frames
    FRAME *Frames[1];                   // the frames
} ANIMATION;

//
// A FACE.DAT block
//

typedef struct _BLOCK {
    OBJECT Hdr;
    uint32_t SectorAddress;     // address of block (1 sector == 2048 bytes)
    uint32_t SectorCount;       // number of sectors
    int Count;                  // number of objects
    OBJECT *Objects[1];         // the objects
} BLOCK;

//
// The contents of a FACE.DAT file
//

typedef struct _FACE_DAT {
    OBJECT *Objects;                    // list of all linked objects
    int Count;                          // number of blocks
    BLOCK *Blocks[1];                   // the blocks
} FACE_DAT;
